#include <Sodaq_SHT2x.h>


#include <SoftwareSerial.h>
#include <EEPROM.h>
#include <Wire.h>
#include <avr/wdt.h>
#include <Sodaq_SHT2x.h>
#include <avr/interrupt.h>

#define MODE_MANU 0     // fonctionnement pwm en mode manuel
#define MODE_ROSEE 1    // fonctionnement pwm par rapport au point de rosÃƒÂ©e
#define MODE_TETA 2     // fonctionnement pwm par rapport ÃƒÂ Ã‚Â  la sonde de tempÃƒÂ©rature
#define LEN_BUFFER 50  // longueur du buffer de reception Bluetooth (max 255);

const String nom_module = "CAAPWM_SylvainV1"; // nom du module bluetooth; maximum 20 carateres; pas d'espace, pas de caractere accentuee
const String code_module = "1357";            // code d'identification module bluetooth
volatile byte msi = 0;                        // valeur ms pour interuption
volatile int timer1_counter = 0;             // compteur tempo pwm

int teta_ext;           // tempÃƒÂ©rature extÃƒÂ©rieure (toute tempÃƒÂ©rature en 10eme)
int hygro_ext;          // hygromÃƒÂ©trie extÃƒÂ©rieure (0 ÃƒÂ  100)
int pts_rosee;          // point de rosÃƒÂ©e calculÃƒÂ©
int teta1;              // tempÃƒÂ©rature sonde 1
int teta2;              // tempÃƒÂ©rature sonde 2
int tension_batterie ;  // tension de la batterie en mv
byte out_pwm1;          // sortie pwm voie 1 sans correction si mode manu
byte out_pwm2;          // sortie pwm voie 1 sans correction si mode manu
byte rout_pwm1;          // valeur de sortie rÃƒÂ©ÃƒÂ©lle corrigÃƒÂ©e en tension pwm voie 1
byte rout_pwm2;          // valeur de sortie rÃƒÂ©ÃƒÂ©lle corrigÃƒÂ©e en tension pwm voie 2

typedef struct __attribute__((__packed__)) {
  byte mode1;                 // mode de fonctionnement voie 1
  byte mode2;                 // mode de fonctionnement voie 2
  byte cons_pwm_manu1;        // consigne pwm mode manuel voie 1
  byte cons_pwm_manu2;        // consigne pwm mode manuel voie 2
  byte cons_max_pwm_v1;       // consigne maximum pwm mode 1 ou 2 voie 1
  byte cons_max_pwm_v2;       // consigne maximum pwm mode 1 ou 2 voie 2
  int cons_delta_m1_v1;       // consigne delta dÃƒÂ©but enclenchement chauffe par rapport au point de rosÃƒÂ©, voie 1
  int cons_delta_m1_v2;       // consigne delta dÃƒÂ©but enclenchement chauffe par rapport au point de rosÃƒÂ©, voie 2
  int cons_delta_m2_v1;       // consigne delta entre tempÃƒÂ©rature ambiante et tempÃƒÂ©rature collier, voie 1
  int cons_delta_m2_v2;       // consigne delta entre tempÃƒÂ©rature ambiante et tempÃƒÂ©rature collier, voie 2
  char calib1;                // calibration sonde collier voie 1
  char calib2;                // calibration sonde collier voie 2
  int crc16;                  // crc
}  sauv_s;

sauv_s sauv;      // initialisation de la structure


// dÃƒÂ©finition des broches d'E/S
SoftwareSerial btSerial(10, 11); // D10 = RX, D11 = TX
const int pinBouton = 12;
const int pinPWM1 = 5;
const int pinPWM2 = 6;
const int pinTeta1 = 0;
const int pinTeta2 = 1;
const int pinVbat = 3;

void setup() {

  // Initialisation du watchdog
  wdt_enable(WDTO_4S);
  wdt_reset();

  // Initialisation des broches d'E/S ainsi que les pÃƒÂ©riphÃƒÂ©riques
  pinMode(LED_BUILTIN, OUTPUT);
  pinMode(pinBouton, INPUT_PULLUP);
  pinMode(pinPWM1, OUTPUT);
  pinMode(pinPWM2, OUTPUT);
  Serial.begin(9600);
  btSerial.begin(9600);
  Wire.begin();
  analogReference(EXTERNAL);


  // configuration interruptions pour generations pwm "doux" à 100 Hz
  noInterrupts();           // Arrete interruptions
  TCCR1A = 0;
  TCCR1B = 0;
  timer1_counter = 64000;
  TCNT1 = timer1_counter;
  TCCR1B |= (1 << CS10);
  TIMSK1 |= (1 << TOIE1);
  interrupts();             // reprise interruptions
}



void loop() {
  byte b;
  unsigned long delay_task_led = 0;  // delai pour clignotement de la led
  unsigned long delay_task_sht = 0;  // delai pour execution tache sht21
  unsigned long delay_task_ana = 0;  // delai pour execution des taches d'aquisition analogiques

  byte led_flip = 0;                // etat de la led
  double t, hr, tr;
  wdt_reset();
  delay(500);

  // verification si demande de reprogrammation du module bluetooh (appuie BP)
  if (digitalRead(pinBouton) == 0) {
    delay(100);
    if (digitalRead(pinBouton) == 0) {
      digitalWrite(LED_BUILTIN, HIGH);
      BLUE_ChangeConfig();
      while (1) {
        wdt_reset();
        digitalWrite(LED_BUILTIN, HIGH);
        delay(50);
        digitalWrite(LED_BUILTIN, LOW);
        delay(50);
      }
    }
  }

  // chargement des parametres de l'eeprom (variables sauvegardees)
  if (EPP_Read() == false) EPP_SetDefault();

  // execute les routines d'aquisition analogique pour remplir le tableau de moyennage
  for (b = 0; b < 5; b++) {
    ANA_Get_Teta1();
    ANA_Get_Teta2();
    ANA_Get_Vbat();
  }


  // boucle de programme principale
  while (1) {

    wdt_reset();

    // tache clignotement led
    if (millis() > delay_task_led) {
      delay_task_led = millis() + 500;
      led_flip = !led_flip;
      digitalWrite(LED_BUILTIN, led_flip);
    }

    // acquisition hygro, tempÃƒÂ©rature, calcul du point de rosÃƒÂ©e
    if (millis() > delay_task_sht) {
      delay_task_sht = millis() + 1800;
      t = SHT2x.GetTemperature();
      hr = SHT2x.GetHumidity();
      teta_ext = t * 10;
      hygro_ext = hr;
      tr = dewPoint(t, hr);
      pts_rosee = tr * 10;
    }

    if (millis() > delay_task_ana) {
      delay_task_ana = millis() + 1000;
      // acquisition des sondes de tempÃƒÂ©ratures externes et filtrage (moyennage)
      ANA_Get_Teta1();
      ANA_Get_Teta2();
      // acquisition tension de la batterie et moyennage
      ANA_Get_Vbat();
    }

    // tache bluetooth
    BLUE_Process_Buffer();

    // tache de regulation
    Kernel_regul();

  }
}


void Kernel_regul () {

  float coef_batterie;
  float p_calcule;
  float f_vbat;
  float f_tmp = 0;
  
  wdt_reset();
  // calcul coefficient correcteur de batterie
  // valeur fictive de référence = 12 V , R = 10 ohms, ce qui donne P = 14.4W
  f_vbat = (float)(tension_batterie) / 10;
  p_calcule = f_vbat * ( f_vbat / 10);
  coef_batterie = 14.4 / p_calcule;

  // mode manu voie 1
  if (sauv.mode1 == MODE_MANU) {
    out_pwm1 = sauv.cons_pwm_manu1;
    f_tmp = out_pwm1;
    f_tmp = f_tmp * coef_batterie;
    if (f_tmp > 100) f_tmp = 100;
    if (f_tmp < 0) f_tmp = 0;
    rout_pwm1 = f_tmp;
  }

  // mode manu voie 2
  if (sauv.mode2 == MODE_MANU) {
    out_pwm2 = sauv.cons_pwm_manu2;
    f_tmp = out_pwm2;
    f_tmp = f_tmp * coef_batterie;
    if (f_tmp > 100) f_tmp = 100;
    if (f_tmp < 0) f_tmp = 0;
    rout_pwm2 = f_tmp ;
  }

  // mode rosée voie 1
  if (sauv.mode1 == MODE_ROSEE) {
    f_tmp = (1 - ( (teta_ext - pts_rosee) / (float)(sauv.cons_delta_m1_v1))) * 100 ;
    f_tmp = f_tmp * coef_batterie;
    if (f_tmp > 100) f_tmp = 100;
    if (f_tmp > sauv.cons_max_pwm_v1) f_tmp = sauv.cons_max_pwm_v1;
    if (f_tmp < 0) f_tmp = 0;
    rout_pwm1 = f_tmp ;
    out_pwm1 = f_tmp;
  }

  // mode rosée voie 2
  if (sauv.mode2 == MODE_ROSEE) {
    f_tmp = (1 - ( (teta_ext - pts_rosee) / (float)(sauv.cons_delta_m1_v2))) * 100 ;
    f_tmp = f_tmp * coef_batterie;
    if (f_tmp > 100) f_tmp = 100;
    if (f_tmp > sauv.cons_max_pwm_v2) f_tmp = sauv.cons_max_pwm_v2;
    if (f_tmp < 0) f_tmp = 0;
    rout_pwm2 = f_tmp ;
    out_pwm2 = f_tmp;
  }

  // mode teta collier voie 1
  if (sauv.mode1 == MODE_TETA) {
    f_tmp = (1 - ( (teta1 - sauv.cons_delta_m2_v1) / (float)(15))) * 100 ;
    f_tmp = f_tmp * coef_batterie;
    if (f_tmp > 100) f_tmp = 100;
    if (f_tmp > sauv.cons_max_pwm_v1) f_tmp = sauv.cons_max_pwm_v1;
    if (f_tmp < 0) f_tmp = 0;
    rout_pwm1 = f_tmp;
    out_pwm1 = f_tmp;
  }

  // mode teta collier voie 2
  if (sauv.mode2 == MODE_TETA) {
    f_tmp = (1 - ( (teta2 - sauv.cons_delta_m2_v2) / (float)(15))) * 100 ;
    f_tmp = f_tmp * coef_batterie;
    if (f_tmp > 100) f_tmp = 100;
    if (f_tmp > sauv.cons_max_pwm_v2) f_tmp = sauv.cons_max_pwm_v2;
    if (f_tmp < 0) f_tmp = 0;
    rout_pwm2 = f_tmp;
    out_pwm2 = f_tmp;
  }
}


// calcul du point de rosee
double dewPoint(double celsius, double humidity) {
  const double a = 17.27;
  const double b = 237.7;
  double temp = (a * celsius) / (b + celsius) + log(humidity * 0.01);
  return (b * temp) / (a - temp);
}


SIGNAL(TIMER1_OVF_vect)
{
  TCNT1 = timer1_counter;   // rechargement compteur
  if (msi >= rout_pwm1) {
    digitalWrite(pinPWM1, LOW);
  }
  else {
    digitalWrite(pinPWM1, HIGH);
  }
  if (msi >= rout_pwm2) {
    digitalWrite(pinPWM2, LOW);
  }
  else {
    digitalWrite(pinPWM2, HIGH);
  }
  if (msi++ > 98)  msi = 0;
}



